# Sample code: Get frame from webcam, run face detection and template extraction on the frame, then display the frame
# Note: you will need to have the opencv-python module installed

import cv2
import os
import sys
import tfsdk
from threading import Thread, Lock
import time

# Utility class for grabbing RTSP stream frames
# Runs in alternate thread to ensure we always have the latest frame from our RTSP stream
class VideoController:
    def __init__(self):
        self.mutex = Lock()

        # Use the default camera (TODO: Can change the camera source, for example to an RTSP stream)
        self.cap = cv2.VideoCapture(0)
        if (self.cap.isOpened()== False): 
            print("Error opening video stream")
            os._exit(1)

        # Launch a thread to grab the newest frame from the frame buffer
        self.thread = Thread(target = self.rtspThreadFunc)
        self.thread.daemon = True
        self.thread.start()

    def rtspThreadFunc(self):
        fps = self.cap.get(cv2.CAP_PROP_FPS)
        sleepTimeSeconds = 1 / fps

        while True:
            # Sleep otherwise we won't give the main thread a chance to acquire the mutex
            time.sleep(0.001)
            
            # Grab a frame from the frame buffer, discard the return value
            # This will discard built up frames in the buffer to ensure we only process the latest frame to remove any delay
            self.mutex.acquire()
            self.cap.grab()
            self.mutex.release()

    # Returns true if a frame was grabbed
    def grabFrame(self):
        self.mutex.acquire()
        ret, image = self.cap.retrieve()
        self.mutex.release()
        return ret, image 

def draw_label(image, point, label,
               font=cv2.FONT_HERSHEY_SIMPLEX,
               font_scale=1.0, thickness=1):
    size = cv2.getTextSize(label, font, font_scale, thickness)[0]
    x_label, y_label = point
    cv2.rectangle(
        image,
        (x_label, y_label - size[1] - 10),
        (x_label + size[0], y_label),
        (194,134,58),
        cv2.FILLED)

    cv2.putText(
        image, label.capitalize(), (x_label, y_label - 5), font, font_scale,
        (255, 255, 255), thickness, cv2.LINE_AA)


def draw_rectangle(frame, bounding_box):

    # Draw the rectangle on the frame
    cv2.rectangle(frame,
                  (int(bounding_box.top_left.x), int(bounding_box.top_left.y)),
                  (int(bounding_box.bottom_right.x), int(bounding_box.bottom_right.y)), (194,134,58), 3)
        


# Create instance of the VideoController class
videoController = VideoController()

options = tfsdk.ConfigurationOptions()
# Can set configuration options here
# ex:
options.smallest_face_height = 40 # https://reference.trueface.ai/cpp/dev/latest/usage/general.html#_CPPv4N8Trueface20ConfigurationOptions18smallestFaceHeightE
# options.fd_mode = tfsdk.FACEDETECTIONMODE.VERSATILE
# options.fd_filter = tfsdk.FACEDETECTIONFILTER.BALANCED
options.fr_model = tfsdk.FACIALRECOGNITIONMODEL.LITE # TODO: Use the full model for improved accuracy
# TODO: If you have a NVIDIA gpu, then enable the enableGPU flag (you will require a GPU specific token for this).
# options.enable_GPU = True 

options.dbms = tfsdk.DATABASEMANAGEMENTSYSTEM.NONE # The data will not persist after the application terminates

# To have the collection data persist after application termination, use the SQLITE or POSTGRESQL backend option
# options.dbms = tfsdk.DATABASEMANAGEMENTSYSTEM.SQLITE
# options.dbms = tfsdk.DATABASEMANAGEMENTSYSTEM.POSTGRESQL

sdk = tfsdk.SDK(options)

# TODO: export your license token as TRUEFACE_TOKEN environment variable
is_valid = sdk.set_license(os.environ['TRUEFACE_TOKEN'])
if (is_valid == False):
    print("Invalid License Provided")
    quit()

# If we are using a database backend, need to first call create_database_connection
# Since we are using DATABASEMANAGEMENTSYSTEM.NONE this is not necessary

# ex. If using POSTGRESQL backend...
# res = sdk.create_database_connection("host=localhost port=5432 dbname=my_database user=postgres password=admin")
# if (res != tfsdk.ERRORCODE.NO_ERROR):
#   print("Unable to create database connection")
#   quit()

# ex. If using SQLITE backend...
# res = sdk.create_database_connection("my_database.db")
# if (res != tfsdk.ERRORCODE.NO_ERROR):
#   print("Unable to create database connection")
#   quit()

# Create a new collection
res = sdk.create_load_collection("my_collection")
if (res != tfsdk.ERRORCODE.NO_ERROR):
    print("Unable to create collection")
    quit()

# Some identities to populate into our collection
image_identities = [
    ("../images/brad_pitt_2.jpg", "Brad Pitt"),
    ("../images/brad_pitt_3.jpg", "Brad Pitt"), # Can add the same identity more than once
    ("../images/tom_cruise_1.jpg", "Tom Cruise")
]

for path, identity in image_identities:
    print("Processing image:", path, "with identity:", identity)
    # Generate a template for each image
    res = sdk.set_image(path)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to set image at path:", path)
        continue

    # Detect the largest face in the image
    found, faceBoxAndLandmarks = sdk.detect_largest_face()
    if found == False:
        print("No face detected in image:", path)
        continue

    # We want to only enroll high quality images into the database / collection
    # For more information, refer to the section titled "Selecting the Best Enrollment Images"
    # https://reference.trueface.ai/cpp/dev/latest/py/identification.html

    # Therefore, ensure that the face height is at least 100px
    faceHeight = faceBoxAndLandmarks.bottom_right.y - faceBoxAndLandmarks.top_left.y
    print("Face height:", faceHeight, "pixels")

    if faceHeight < 100:
        print("The face is too small in the image for a high quality enrollment, not enrolling")
        continue

    # Get the aligned chip so we can compute the image quality
    face = sdk.extract_aligned_face(faceBoxAndLandmarks)

    # Compute the image quality score
    res, quality = sdk.estimate_face_image_quality(face)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("There was an error computing the image quality")
        continue

    # Ensure the image quality is above a threshold (TODO: adjust this threshold based on your use case).
    print("Face quality:", quality)
    if quality < 0.8:
        print("The image quality is too poor for enrollment, not enrolling")
        continue

    # As a final check, we can check the orientation of the head and ensure that it is facing forward
    # To see the effect of yaw and pitch on match score, refer to: https://reference.trueface.ai/cpp/dev/latest/py/face.html#tfsdk.SDK.estimate_head_orientation

    res, yaw, pitch, roll = sdk.estimate_head_orientation(faceBoxAndLandmarks)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to compute head orientation")

    yaw_deg = yaw * 180 / 3.14
    pitch_deg = yaw * 180 / 3.14

    if abs(yaw_deg) > 50:
        print("Enrollment image has too extreme a yaw, not enrolling")
        continue

    if abs(pitch_deg) > 35:
        print("Enrollment image has too extreme a pitch, not enrolling")
        continue

    # Finally ensure the user is not wearing a mask
    error_code, mask_label, score = sdk.detect_mask(faceBoxAndLandmarks)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to run mask detection")
        continue

    if (mask_label == tfsdk.MASKLABEL.MASK):
        print("Please choose a image without a mask for enrollment, not enrolling")
        continue

    # Now that we have confirmed the images are high quality, generate a template from that image
    res, faceprint = sdk.get_face_feature_vector(face)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("There was an error generating the faceprint")
        continue

    # Enroll the feature vector into the collection
    res, UUID = sdk.enroll_faceprint(faceprint, identity)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to enroll feature vector")
        continue

    # TODO: Can store the UUID for later use
    print("Success, enrolled template with UUID:", UUID)
    print("--------------------------------------------")


filename = "my_window"

while(True):
    ret, frame = videoController.grabFrame()
    if ret == False:
        continue

    # Set the image using the frame buffer. OpenCV stores images in BGR format
    res = sdk.set_image(frame, frame.shape[1], frame.shape[0], tfsdk.COLORCODE.bgr)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to set frame.")
        cv2.imshow(filename, frame)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break
        continue

    faceboxes = sdk.detect_faces()
    if not faceboxes or len(faceboxes) == 0:
        cv2.imshow(filename, frame)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break
        continue

    # Run 1:N search for all extracted faces
    for facebox in faceboxes:
        # Extract the feature vector
        # We do not need to check the quality of the probe face templates
        # We mainly want to ensure that the enrollment templates are high quality, 
        # This is less of a concern with probe templates 
        res, faceprint = sdk.get_face_feature_vector(facebox)
        if (res != tfsdk.ERRORCODE.NO_ERROR):
            print("skipping facebox")
            continue

        res, match_found, candidate = sdk.identify_top_candidate(faceprint, threshold=0.3)
        draw_rectangle(frame, facebox)

        if (res != tfsdk.ERRORCODE.NO_ERROR or not match_found):
            continue
        else:
            draw_label(frame,
                       (int(facebox.top_left.x), int(facebox.top_left.y)),
                       "{} {}%".format(
                           candidate.identity,
                           int(candidate.match_probability*100)))


    cv2.imshow(filename, frame)
    if cv2.waitKey(1) & 0xFF == ord('q'):
        break


# When everything done, release the capture
cap.release()
cv2.destroyAllWindows()




